(* function for feeding strings to buffer and retrieving whole lines *)
let add_buf_str buf str length =
  let rec add_buf_at buf ofs = 
    let rec find_eol at =
      if at < length then
        if str.[at] = '\n' then
          Some at
        else
          find_eol (at + 1)
       else
         None
    in
      match find_eol ofs with
       | None -> 
         if length <> ofs then
           Buffer.add_substring buf str ofs (length - ofs);
         []
       | Some at ->
         Buffer.add_substring buf str ofs (at - ofs);
         let str = Buffer.contents buf in
           Buffer.clear buf;
           str::add_buf_at buf (at + 1)
  in
    add_buf_at buf 0

let main () = 
  (* the default event loop *)
  let loop = Ev.default_loop [] in

  (* intermediate storage for incoming bytes *)
  let buf = Buffer.create 1024 in

  let exit_io io =
    Printf.printf "Bye!\n%!"; 
    Ev.stop io
  in

  (* a receiving buffer for a single message *)
  let str_buf = String.create 1024 in
  (* start io immediately after creating it *)
  let io_wait = Ev.io_init Unix.stdin [Ev.READ] Ev.no_cb in
  let timeout_wait = Ev.timer_init 5.0 5.0 Ev.no_cb in
    Ev.set_cb io_wait (
      (* use io parameter for an easy mechanism to stop itself *)
      (* revents parameter is ignored *)
      fun io _ ->
        (* receive data from the peer.. (stdin in this case) *)
        try
          let n = Unix.read Unix.stdin str_buf 0 (String.length str_buf) in
            (* bail out when there is an EOF or an error (doesn't handle EAGAIN) *)
            if n <= 0 then exit_io io
            else
              ((* reset timer *)
                Ev.timer_again timeout_wait;
                (* output received messages *)
                List.iter
                  (fun str -> 
                     Printf.printf "Received: %s\n%!" str;
                     if str = "bye" then
		       begin
			 Ev.stop io;
			 Ev.stop timeout_wait;
		       end
                  )
                  (add_buf_str buf str_buf n)
              )
        with 
          | Unix.Unix_error ((Unix.EAGAIN | Unix.EINTR), _, _) -> ()
          | Unix.Unix_error _ -> exit_io io
    );
    Ev.set_cb timeout_wait (
      fun io revents ->
        Printf.printf "Timeout! %s\n%!" (String.concat " " (List.map Ev.string_of_revent revents));
	Ev.unloop (Ev.get_loop io) Ev.UnloopOne;
    );

    Ev.start loop io_wait;
    Ev.start loop timeout_wait;

    (* set the descriptor to a non-blocking mode just in case (certain platforms may produce
       spurious events) *)
    Unix.set_nonblock Unix.stdin;
    (* start it! *)
    Ev.loop loop [];
    Unix.clear_nonblock Unix.stdin

let () = main ()
